@@ -92,17 +92,23 @@ use Benchmark ':hireswallclock';
 
 use Text::LevenshteinXS;
 
+use Data::Dumper;
+
 #use Text::WagnerFischer qw(distance);
 
 Log::Log4perl->easy_init( { level => $INFO, layout => '%-5p - %m%n' } );
 
+# script arguments
+my (
+	$owlfile,   $obofile,   $targetfile, $resultfile,
+	$obotarget, $owltarget, $meshfile
+);
+my $flat_header = undef;
+my @flat_header;
+
 sub main() {
 
 	# initalize
-	my (
-		$owlfile,   $obofile,   $targetfile, $resultfile,
-		$obotarget, $owltarget, $meshfile
-	);
 
 	GetOptions(
 		"o|obofile=s"  => \$obofile,
@@ -134,7 +140,7 @@ sub main() {
 
 	#print Dumper( \%term );
 
-	align( $data, $ontology, $resultfile );
+	process_and_write( $data, $ontology, $resultfile );
 }
 
 sub usage() {
@@ -219,15 +225,14 @@ on non-alphanumerics.
 sub normalise($) {
 	my $word = shift;
 
-	# trim
-	$word =~ s/^\s+//;
-	$word =~ s/\s+$//;
-
 	# lowercase has to be done before compare
 	$word = lc($word);
 
-	# Split on non-alphanumeric chars, sort words, join
-	$word = join( q{}, sort { $a cmp $b } ( split /[[:^alnum:]]/, $word ) );
+	# split on non-alphanumeric chars, sort words, join
+	#$word = join( q{}, sort { $a cmp $b } ( split /[[:^alnum:]]/, $word ) );
+
+	# sort characters alphabetically
+	$word = join '', sort split( '', $word );
 }
 
 =item align()
@@ -237,23 +242,45 @@ the results into a file.
 
 =cut
 
-sub align($$$) {
+sub process_and_write($$$) {
 	my ( $data, $ontology, $file ) = @_;
 	open my $fh_out, '>', $file;
-	print $fh_out "ID\tOE_VALUE\tONTOLOGY_TERM\tACCESSION\tSIMILARITY%\n";
 	$fh_out->autoflush(1);
 
-	# process
+	# write header
+	if ( defined $obotarget || defined $owltarget ) {
+		print $fh_out "ID\tOE_VALUE\t";
+	}
+	else {
+		print $fh_out $flat_header[0] . "\t";
+	}
+	print $fh_out "ONTOLOGY_TERM\tACCESSION\tSIMILARITY%";
+	print $fh_out "\t$flat_header[1]" if defined $flat_header;
+	print $fh_out "\n";
+
 	my $c;
 	my $t0 = new Benchmark;
 
 	for my $id ( keys %$data ) {
 		$c++;
 		my $label = $data->{$id}->{label};
-		print $fh_out $id . "\t" . process( $ontology, $label );
 
+		# do not output id for flat files as it's same as label
+		print $fh_out $id . "\t"
+		  if defined $obotarget || defined $owltarget;
+
+		# output match info
+		print $fh_out find_match( $ontology, $label );
+
+		# output unprocessed columns back
+		print $fh_out "\t" . $data->{$id}->{ragged_end};
+
+		# line ending
+		print $fh_out "\n";
+
+		# this only happens for owl or obo targets
 		for my $synonym ( @{ $data->{$id}->{synonyms} } ) {
-			print $fh_out $id . "\t" . process( $ontology, $synonym );
+			print $fh_out $id . "\t" . find_match( $ontology, $synonym ) . "\n";
 			$c++;
 		}
 		INFO "Processed " . $c
@@ -305,10 +332,31 @@ sub parseFlat($) {
 
 	open my $fh_in, '<', $file;
 
+	# parse header
+	$flat_header = <$fh_in>;
+	chomp($flat_header);
+	@flat_header = $flat_header =~ /(.*?)\t(.*)/;
+	INFO "Using first line as header <$flat_header>";
+	INFO "Using first column <$flat_header[0]> to match terms";
+
 	# load input
 	while (<$fh_in>) {
 		chomp;
-		$term->{$_}->{label} = $_;
+
+		# preserve existing columns in the file
+		my ( $label, $ragged_end ) = /(.*?)\t(.*)/;
+
+		# trim
+		$label =~ s/^\s+//;
+		$label =~ s/\s+$//;
+
+		# drop trailing quotation marks (excel artefact?)
+		$label =~ s/^"+//;
+		$label =~ s/"+$//;
+
+		WARN "Duplicated <$label>" if exists $term->{$label};
+		$term->{$label}->{label}      = $label;
+		$term->{$label}->{ragged_end} = $ragged_end;
 	}
 
 	close $fh_in;
@@ -318,7 +366,7 @@ sub parseFlat($) {
 	return $term;
 }
 
-=item parseFlat()
+=item parseOBO()
 
 Custom OBO parser.
 
@@ -359,7 +407,7 @@ sub parseOBO($) {
 	return $term;
 }
 
-=item parseFlat()
+=item parseOWL()
 
 Custom OWL parser.
 
@@ -388,24 +436,21 @@ sub parseOWL($) {
 	return $term;
 }
 
-=item process()
+=item find_match()
 
 A wrapper around the calculate_distance function. Specifies the
-similarity metric to be used, in this case Text::LevenshteinXS::distance
-and normalises the some of the input.
+similarity metric to be used, in this case Text::LevenshteinXS::distance.
 
 Outputs a single line in the output file.
 
 =cut
 
-sub process($$) {
-	my ( $term, $term_to_match ) = @_;
+sub find_match($$) {
+	my ( $ontology, $term_to_match ) = @_;
 
-	#Text::WagnerFischer::distance( [ 0, 1, 2 ]
-	# term hash was already normalised
 	my $distance = calculate_distance(
-		$term,
-		normalise($term_to_match),
+		$ontology,
+		$term_to_match,
 		sub($$) {
 			my ( $word1, $word2 ) = @_;
 			return Text::LevenshteinXS::distance( $word1, $word2 );
@@ -415,19 +460,17 @@ sub process($$) {
 	my $matched_term = $distance->{term};
 	my $type         = $distance->{type};
 	my $matched_acc  = $distance->{acc};
-	my $min_distance = $distance->{dist};
-	my $similarity   = int(
-		( ( length($term_to_match) - $min_distance ) / length($term_to_match) )
-		* 100 );
+	my $similarity   = $distance->{sim};
 
-	$matched_term = "SYN: $matched_term OF: " . $term->{$matched_acc}->{label}
+	$matched_term =
+	  "SYN: $matched_term OF: " . $ontology->{$matched_acc}->{label}
 	  if $type eq "synonym";
 
 	my $output_str =
 	    $term_to_match . "\t"
 	  . $matched_term . "\t"
 	  . $matched_acc . "\t"
-	  . $similarity . "\n";
+	  . $similarity;
 
 	#INFO $output_str;
 	return $output_str;
@@ -437,22 +480,26 @@ sub process($$) {
 
 Finds the best match for the supplied term in the ontology
 using the supplied anonymous distance function defined in
-process().
+find_match().
 
 =cut
 
 sub calculate_distance($$&) {
-	my ( $term, $term_to_match, $distance_function ) = @_;
+	my ( $ontology, $term_to_match, $distance_function ) = @_;
 	my $matched_term;
 	my $matched_acc;
 	my $type;
 	my $min_distance = undef;
+	my $similarity;
 
-	for my $id ( keys %$term ) {
-		my $label            = $term->{$id}->{label};
-		my $normalised_label = $term->{$id}->{normalised_label};
+	# ontology hash was already prenormalised earlier
+	my $normalised_term = normalise($term_to_match);
+
+	for my $id ( keys %$ontology ) {
+		my $label            = $ontology->{$id}->{label};
+		my $normalised_label = $ontology->{$id}->{normalised_label};
 		my $distance =
-		  $distance_function->( $term_to_match, $normalised_label );
+		  $distance_function->( $normalised_term, $normalised_label );
 		if ( !( defined $min_distance ) || $distance < $min_distance ) {
 			$min_distance = $distance;
 			$matched_term = $label;
@@ -460,11 +507,11 @@ sub calculate_distance($$&) {
 			$type         = "label";
 		}
 
-		for my $synonym ( keys %{ $term->{$id}->{normalised_syns_hash} } ) {
+		for my $synonym ( keys %{ $ontology->{$id}->{normalised_syns_hash} } ) {
 			my $normalised_syn =
-			  $term->{$id}->{normalised_syns_hash}->{$synonym};
+			  $ontology->{$id}->{normalised_syns_hash}->{$synonym};
 			my $distance =
-			  $distance_function->( $term_to_match, $normalised_syn );
+			  $distance_function->( $normalised_term, $normalised_syn );
 			if ( !( defined $min_distance ) || $distance < $min_distance ) {
 				$min_distance = $distance;
 				$matched_term = $synonym;
@@ -474,10 +521,29 @@ sub calculate_distance($$&) {
 		}
 	}
 
+	if ( lc($matched_term) eq lc($term_to_match) ) {    # perfect match
+		$similarity = 100;
+	}
+	else {
+
+		# normalise the minimal LD distance
+		$similarity = int(
+			(
+				( length($term_to_match) - $min_distance ) /
+				  length($term_to_match)
+			) * 100
+		);
+		if ( $similarity == 100 ) {
+			WARN
+"Possible anagram mismatch, mapping <$term_to_match> to <$matched_term> at at 99% similarity";
+			$similarity = 99;    # make sure it's eyballed at some point
+		}
+	}
+
 	return {
 		term => $matched_term,
 		acc  => $matched_acc,
-		dist => $min_distance,
+		sim  => $similarity,
 		type => $type,
 	};
 }
